"""
Declares enumerations for various model properties.
"""
from bokeh.core.enums import enumeration

ace_themes = enumeration(
    'ambiance', 'chaos', 'chrome', 'cloud9_day', 'cloud9_night', 'clouds',
    'clouds_midnight', 'cobalt', 'crimson_editor', 'dawn', 'dracula',
    'dreamweaver', 'eclipse', 'github', 'github_dark',
    'github_light_default', 'gob', 'gruvbox', 'idle_fingers',
    'iplastic', 'katzenmilch', 'kr_theme', 'kuroir', 'merbivore',
    'merbivore_soft', 'mono_industrial', 'monokai', 'nord_dark',
    'one_dark', 'pastel_on_dark', 'solarized_dark', 'solarized_light',
    'sqlserver', 'terminal', 'textmate', 'tomorrow', 'tomorrow_night',
    'tomorrow_night_blue', 'tomorrow_night_bright',
    'tomorrow_night_eighties', 'twilight', 'vibrant_ink', 'xcode',
)



import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel import Feed
from panel.layout.spacer import Spacer
from panel.tests.util import serve_component, wait_until

pytestmark = [pytest.mark.ui, pytest.mark.flaky(max_runs=3)]

ITEMS = 100  # 1000 items make the CI flaky

def test_feed_load_entries(page):
    feed = Feed(*list(range(ITEMS)), height=250)
    serve_component(page, feed)

    feed_el = page.locator(".bk-panel-models-feed-Feed")

    bbox = feed_el.bounding_box()
    assert bbox["height"] == 250

    expect(feed_el).to_have_class("bk-panel-models-feed-Feed scroll-vertical")

    children_count = feed_el.locator('.bk-panel-models-markup-HTML').count()
    assert 50 <= children_count <= 65

    # Now scroll to somewhere down
    feed_el.evaluate('(el) => el.scrollTo({top: 100})')
    children_count = feed_el.locator('.bk-panel-models-markup-HTML').count()
    assert 50 <= children_count <= 65

    # Now scroll to top
    feed_el.evaluate('(el) => el.scrollTo({top: 0})')
    wait_until(lambda: feed_el.locator('.bk-panel-models-markup-HTML').count() >= 50, page)

def test_feed_view_latest(page):
    feed = Feed(*list(range(ITEMS)), height=250, view_latest=True)
    serve_component(page, feed)

    feed_el = page.locator(".bk-panel-models-feed-Feed")

    bbox = feed_el.bounding_box()
    assert bbox["height"] == 250

    expect(feed_el).to_have_class("bk-panel-models-feed-Feed scroll-vertical")

    # Assert scroll is not at 0 (view_latest)
    wait_until(lambda: feed_el.evaluate('(el) => el.scrollTop') > 0, page)

    wait_until(lambda: int(page.locator('pre').last.inner_text() or 0) > 0.9 * ITEMS, page)


def test_feed_view_scroll_to_latest(page):
    feed = Feed(*list(range(ITEMS)), height=250)
    serve_component(page, feed)

    feed_el = page.locator(".bk-panel-models-feed-Feed")

    bbox = feed_el.bounding_box()
    assert bbox["height"] == 250

    expect(feed_el).to_have_class("bk-panel-models-feed-Feed scroll-vertical")

    # Assert scroll is not at 0 (view_latest)
    wait_until(lambda: feed_el.evaluate('(el) => el.scrollTop') == 0, page)

    feed.scroll_to_latest()

    wait_until(lambda: int(page.locator('pre').last.inner_text() or 0) > 0.9 * ITEMS, page)


def test_feed_scroll_to_latest_disabled_when_limit_zero(page):
    """Test that scroll_to_latest is disabled when scroll_limit = 0"""
    feed = Feed(*list(range(ITEMS)), height=250)
    serve_component(page, feed)

    feed_el = page.locator(".bk-panel-models-feed-Feed")
    initial_scroll = feed_el.evaluate('(el) => el.scrollTop')

    # Try to scroll to latest
    feed.scroll_to_latest(scroll_limit=0)

    # Verify scroll position hasn't changed
    final_scroll = feed_el.evaluate('(el) => el.scrollTop')
    assert initial_scroll == final_scroll, "Scroll position should not change when limit is 0"


def test_feed_scroll_to_latest_always_when_limit_null(page):
    """Test that scroll_to_latest always triggers when scroll_limit is null"""
    feed = Feed(*list(range(ITEMS)), height=250)
    serve_component(page, feed)

    wait_until(lambda: int(page.locator('pre').last.inner_text() or 0) < 0.9 * ITEMS, page)
    feed.scroll_to_latest(scroll_limit=None)
    wait_until(lambda: int(page.locator('pre').last.inner_text() or 0) > 0.9 * ITEMS, page)


def test_feed_scroll_to_latest_within_limit(page):
    """Test that scroll_to_latest only triggers within the specified limit"""
    feed = Feed(
        Spacer(styles=dict(background='red'), width=200, height=200),
        Spacer(styles=dict(background='green'), width=200, height=200),
        Spacer(styles=dict(background='blue'), width=200, height=200),
        auto_scroll_limit=0, height=420
    )
    serve_component(page, feed)

    feed_el = page.locator(".bk-panel-models-feed-Feed")

    expect(feed_el).to_have_js_property('scrollTop', 0)

    feed.scroll_to_latest(scroll_limit=100)

    feed.append(Spacer(styles=dict(background='yellow'), width=200, height=200))

    # assert scroll location is still at top
    expect(feed_el.locator('div')).to_have_count(5)
    expect(feed_el).to_have_js_property('scrollTop', 0)

    # scroll to close to bottom
    feed_el.evaluate('(el) => el.scrollTo({top: 200})')
    expect(feed_el).to_have_js_property('scrollTop', 200)

    # assert auto scroll works; i.e. distance from bottom is 0
    feed.append(Spacer(styles=dict(background='yellow'), width=200, height=200))
    feed.scroll_to_latest(scroll_limit=1000)

    def assert_at_bottom():
        assert feed_el.evaluate(
            '(el) => el.scrollHeight - el.scrollTop - el.clientHeight'
        ) == 0
    wait_until(assert_at_bottom, page)


def test_feed_view_scroll_button(page):
    feed = Feed(*list(range(ITEMS)), height=250, scroll_button_threshold=50)
    serve_component(page, feed)

    feed_el = page.locator(".bk-panel-models-feed-Feed")

    # assert scroll button is visible on render
    scroll_arrow = page.locator(".scroll-button")
    expect(scroll_arrow).to_have_class('scroll-button visible')
    expect(scroll_arrow).to_be_visible()

    # click on scroll arrow
    scroll_arrow.click()

    # Assert scroll is not at 0 (view_latest)
    wait_until(lambda: feed_el.evaluate('(el) => el.scrollTop') > 0, page)
    wait_until(lambda: int(page.locator('pre').last.inner_text() or 0) > 50, page)

def test_feed_dynamic_objects(page):
    feed = Feed(height=250, load_buffer=10)
    serve_component(page, feed)

    feed.objects = list(range(ITEMS))

    wait_until(lambda: expect(page.locator('pre').first).to_have_text('0'))
    wait_until(lambda: page.locator('pre').count() > 10, page)

def test_feed_reset_visible_range(page):
    feed = Feed(*list(range(ITEMS)), load_buffer=20, height=50, view_latest=True)
    serve_component(page, feed)

    wait_until(lambda: int(page.locator('pre').last.inner_text() or 0) == 99, page)

    # set objects to 20
    feed.objects = feed.objects[:20]

    # assert view reset
    wait_until(lambda: (
        int(page.locator('pre').last.inner_text() or 0) == 19
    ), page)



import pytest

from panel.layout import Modal, Spacer
from panel.tests.util import serve_component

pytest.importorskip("playwright")

from playwright.sync_api import expect

pytestmark = pytest.mark.ui


def test_modal_close_open(page):
    modal = Modal(
        Spacer(styles=dict(background="red"), width=200, height=200),
        Spacer(styles=dict(background="green"), width=200, height=200),
        Spacer(styles=dict(background="blue"), width=200, height=200),
    )

    serve_component(page, modal)

    modal_locator = page.locator("#pnx_dialog_content")
    expect(modal_locator).to_be_hidden()

    modal.open = True
    expect(modal_locator).to_be_visible()


def test_modal_open_close(page):
    modal = Modal(
        Spacer(styles=dict(background="red"), width=200, height=200),
        Spacer(styles=dict(background="green"), width=200, height=200),
        Spacer(styles=dict(background="blue"), width=200, height=200),
        open=True,
    )

    serve_component(page, modal)

    modal_locator = page.locator("#pnx_dialog_content")
    expect(modal_locator).to_be_visible()

    page.mouse.click(0, 0)
    expect(modal_locator).to_be_hidden()


@pytest.mark.parametrize("show_close_button", [True, False])
def test_modal_show_close_button(page, show_close_button):
    modal = Modal(
        Spacer(styles=dict(background="red"), width=200, height=200),
        Spacer(styles=dict(background="green"), width=200, height=200),
        Spacer(styles=dict(background="blue"), width=200, height=200),
        open=True,
        show_close_button=show_close_button,
    )

    serve_component(page, modal)

    modal_locator = page.locator("#pnx_dialog_content")
    expect(modal_locator).to_be_visible()

    modal_locator = page.locator("#pnx_dialog_close")
    if show_close_button:
        expect(modal_locator).to_be_visible()
    else:
        expect(modal_locator).to_be_hidden()


def test_modal_background_close(page):
    modal = Modal(
        Spacer(styles=dict(background="red"), width=200, height=200),
        Spacer(styles=dict(background="green"), width=200, height=200),
        Spacer(styles=dict(background="blue"), width=200, height=200),
        open=True,
        background_close=False,
    )

    serve_component(page, modal)

    modal_locator = page.locator("#pnx_dialog_content")
    expect(modal_locator).to_be_visible()

    # Should still be visible
    page.mouse.click(0, 0)
    expect(modal_locator).to_be_visible()



from __future__ import annotations

import os
import pathlib
import sys

from bokeh.command.subcommand import Argument, Subcommand

from ..custom import ReactiveESM
from ..io.compile import RED, compile_components, find_module_bundles


def run_compile(
    bundles: dict[pathlib.Path, list[type[ReactiveESM]]],
    build_dir: str | os.PathLike | None = None,
    unminified: bool = False,
    skip_npm: bool = False,
    file_loaders: list[str] | None = None,
    verbose: bool = False
) -> int:
    """
    Runs the compile command on the provided bundles.

    Parameters
    ----------
    bundles : dict[type[ReactiveESM]]
        A list of `ReactiveESM` component classes to compile.
    build_dir : str | os.PathLike, optional
        The directory where the build output will be saved. If None, a
        temporary directory will be used.
    unminified : bool, optional
        If True, minifies the compiled JavaScript bundle.
    skip_npm: bool
        Whether to skip npm install (assumes build_dir is set)
    file_loaders: list[str]
        List of file types (e.g. woff2 or svg) loaders to carry along
    verbose : bool, optional
        If True, prints detailed logs during the compilation process.

    Returns
    -------
    int:
       Count of errors.
    """
    file_loaders = file_loaders or []
    errors = 0
    for bundle, components in bundles.items():
        component_names = '\n- '.join(c.name for c in components)
        print(f"Building {bundle} containing the following components:\n\n- {component_names}\n")  # noqa
        out = compile_components(
            components,
            build_dir=build_dir,
            minify=not unminified,
            outfile=bundle,
            skip_npm=skip_npm,
            file_loaders=file_loaders,
            verbose=verbose,
        )
        if not out:
            errors += 1
    return errors


class Compile(Subcommand):
    ''' Subcommand to generate a new encryption key.

    '''

    #: name for this subcommand
    name = "compile"

    help = "Compiles an ESM component using node and esbuild"

    args = (
        ('modules', Argument(
            metavar = 'DIRECTORY-OR-SCRIPT',
            nargs   = "*",
            help    = "The Python modules to compile. May optionally define a single class.",
            default = None,
        )),
        ('--build-dir', Argument(
            action = 'store',
            type    = str,
            help    = "Where to write the build directory."
        )),
        ('--skip-npm', Argument(
            action  = 'store_true',
            help    = "Whether to skip npm install step (only possible if build_dir with existing node_modules is available)."
        )),
        ('--unminified', Argument(
            action  = 'store_true',
            help    = "Whether to generate unminified output."
        )),
        ('--file-loaders', Argument(
            action  = 'store',
            nargs   = '*',
            help    = "List of file types to include in the compiled bundle.",
        )),
        ('--verbose', Argument(
            action  = 'store_true',
            help    = "Whether to show verbose output. Note when setting --outfile only the result will be printed to stdout."
        )),
        ('--watch', Argument(
            action  = 'store_true',
            help    = 'Whether to watch the input files for changes and recompile the bundle.'
        )),
    )

    def invoke(self, args):
        bundles = {}
        for module in args.modules:
            try:
                module_bundles = find_module_bundles(module)
            except RuntimeError as e:
                print(f'{RED} {e}')  # noqa
                return 1
            if not module_bundles:
                print (  # noqa
                    f'{RED} Could not find any ESM components to compile '
                    f'in {module}, ensure you provided the right module.'
                )
            for bundle, components in module_bundles.items():
                if bundle in bundles:
                    bundles[bundle] += components
                else:
                    bundles[bundle] = components

        errors = run_compile(
            bundles, args.build_dir, args.unminified, args.skip_npm, args.file_loaders, args.verbose
        )
        if args.watch:
            from watchfiles import watch
            paths_to_watch = set()
            for _, components in bundles.items():
                for component in components:
                    mod = sys.modules[component.__module__]
                    mod_path = pathlib.Path(mod.__file__)
                    paths_to_watch.add(mod_path)
                    paths_to_watch.add(mod_path.parent / component._esm_path(compiled='compiling'))
                    for shared in component._esm_shared.values():
                        if isinstance(shared, os.PathLike):
                            paths_to_watch.add(shared)

            for _changes in watch(*paths_to_watch):
                errors = run_compile(
                    bundles, args.build_dir, args.unminified, args.skip_npm, args.file_loaders, args.verbose
                )
        return int(errors>0)



import importlib.util
import subprocess
import sys

import pytest


@pytest.mark.skipif(not importlib.util.find_spec("cryptography"), reason='cryptography is not installed')
def test_oauth_secret_command():
    cmd = [sys.executable, "-m", "panel", "oauth-secret"]
    p = subprocess.Popen(cmd, shell=False, text=True, stdout=subprocess.PIPE)
    p.wait()

    try:
        assert len(p.stdout.read()) == 45
    finally:
        p.stdout.close()



"""
Contains Media panes including renderers for Audio and Video content.
"""
from __future__ import annotations

import os

from base64 import b64encode
from collections.abc import Mapping
from io import BytesIO
from typing import Any, ClassVar

import numpy as np
import param

from ..models import Audio as _BkAudio, Video as _BkVideo
from ..util import isfile, isurl
from .base import ModelPane


class TensorLikeMeta(type):
    """See https://blog.finxter.com/python-__instancecheck__-magic-method/"""
    def __instancecheck__(self, instance):
        numpy_attr = getattr(instance, "numpy", "")
        dim_attr = getattr(instance, "dim", "")
        return (
            bool(numpy_attr) and
            callable(numpy_attr) and
            callable(dim_attr) and
            hasattr(instance, "dtype")
        )

# When support for Python 3.7 is dropped, Torchlike should be implemented as a typing.Protocol
# That would provide type checking and intellisense in editors
class TensorLike(metaclass=TensorLikeMeta):
    """A class similar to torch.Tensor. We don't want to make PyTorch a dependency of this project
    """

class _MediaBase(ModelPane):

    loop = param.Boolean(default=False, doc="""
        Whether the media should loop""")

    time = param.Number(default=0, doc="""
        The current timestamp""")

    throttle = param.Integer(default=250, doc="""
        How frequently to sample the current playback time in milliseconds""")

    paused = param.Boolean(default=True, doc="""
        Whether the media is currently paused""")

    object = param.String(default='', allow_None=True, doc="""
        The media file either local or remote.""")

    volume = param.Number(default=None, bounds=(0, 100), doc="""
        The volume of the media player.""")

    autoplay = param.Boolean(default=False, doc="""
        When True, it specifies that the output will play automatically.
        In Chromium browsers this requires the user to click play once.""")

    muted = param.Boolean(default=False, doc="""
        When True, it specifies that the output should be muted.""")

    _default_mime: ClassVar[str]

    _formats: ClassVar[list[str]]

    _media_type: ClassVar[str]

    _rename: ClassVar[Mapping[str, str | None]] = {
        'sample_rate': None, 'object': 'value'
    }

    _rerender_params: ClassVar[list[str]] = []

    _updates: ClassVar[bool] = True

    __abstract = True

    @classmethod
    def applies(cls, obj: Any) -> float | bool | None:
        if isinstance(obj, str):
            if isfile(obj) and any(obj.endswith('.'+fmt) for fmt in cls._formats):
                return True
            if isurl(obj, cls._formats):
                return True
        if hasattr(obj, 'read'):  # Check for file like object
            return True
        return False

    def _to_np_int16(self, data: np.ndarray) -> np.ndarray:
        dtype = data.dtype

        if dtype in (np.float32, np.float64):
            data = (data * 32768.0).astype(np.int16)

        return data

    def _to_buffer(self, data: np.ndarray | TensorLike):
        if isinstance(data, np.ndarray):
            values = data
        elif isinstance(data, TensorLike):
            values = data.numpy()
        data = self._to_np_int16(values)

        from scipy.io import wavfile
        buffer = BytesIO()
        wavfile.write(buffer, self.sample_rate, data)
        return buffer

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        if 'js_property_callbacks' in msg:
            del msg['js_property_callbacks']
        return msg

    def _transform_object(self, obj: Any) -> dict[str, Any]:
        fmt = self._default_mime
        if obj is None:
            data = b''
        elif isinstance(obj, (np.ndarray, TensorLike)):
            fmt = 'wav'
            buffer = self._to_buffer(obj)
            data = b64encode(buffer.getvalue())
        elif os.path.isfile(obj):
            fmt = obj.split('.')[-1]
            with open(obj, 'rb') as f:
                data = f.read()
            data = b64encode(data)
        elif obj.lower().startswith('http'):
            return dict(object=obj)
        elif not obj or obj == f'data:{self._media_type}/{fmt};base64,':
            data = b''
        else:
            raise ValueError(f'Object should be either path to a {self._media_type} file or numpy array.')
        b64 = f"data:{self._media_type}/{fmt};base64,{data.decode('utf-8')}"
        return dict(object=b64)

_VALID_TORCH_DTYPES_FOR_AUDIO = [
    "torch.short", "torch.int16",
    "torch.half", "torch.float16",
    "torch.float", "torch.float32",
    "torch.double", "torch.float64",
]

_VALID_NUMPY_DTYPES_FOR_AUDIO = [np.int16, np.uint16, np.float32, np.float64]

def _is_1dim_int_or_float_tensor(obj: Any)->bool:
    return (
        isinstance(obj, TensorLike) and
        obj.dim()==1 and
        str(obj.dtype) in _VALID_TORCH_DTYPES_FOR_AUDIO
    )

def _is_1dim_int_or_float_ndarray(obj: Any)->bool:
    return (
        isinstance(obj, np.ndarray) and
        obj.ndim==1 and
        obj.dtype in _VALID_NUMPY_DTYPES_FOR_AUDIO
    )

class Audio(_MediaBase):
    """
    The `Audio` pane displays an audio player given a local or remote audio
    file, a NumPy Array or Torch Tensor.

    The pane also allows access and control over the player state including
    toggling of playing/paused and loop state, the current time, and the
    volume.

    The audio player supports ogg, mp3, and wav files

    If SciPy is installed, 1-dim Numpy Arrays and 1-dim
    Torch Tensors are also supported. The dtype must be one of the following

    - numpy: np.int16, np.uint16, np.float32, np.float64
    - torch: torch.short, torch.int16, torch.half, torch.float16, torch.float, torch.float32,
    torch.double, torch.float64

    The array or Tensor input will be downsampled to 16bit and converted to a wav file by SciPy.

    Reference: https://panel.holoviz.org/reference/panes/Audio.html

    :Example:

    >>> Audio('http://ccrma.stanford.edu/~jos/mp3/pno-cs.mp3', name='Audio')
    """

    object = param.ClassSelector(default='', class_=(str, np.ndarray, TensorLike),
                                 allow_None=True, doc="""
        The audio file either local or remote, a 1-dim NumPy ndarray or a 1-dim Torch Tensor.""")

    sample_rate = param.Integer(default=44100, doc="""
        The sample_rate of the audio when given a NumPy array or Torch tensor.""")

    _bokeh_model = _BkAudio

    _default_mime = 'wav'

    _formats = ['mp3', 'wav', 'ogg']

    _media_type = 'audio'

    @classmethod
    def applies(cls, obj: Any) -> float | bool | None:
        return (super().applies(obj)
            or _is_1dim_int_or_float_ndarray(obj)
            or _is_1dim_int_or_float_tensor(obj)
        )


class Video(_MediaBase):
    """
    The `Video` Pane displays a video player given a local or remote video
    file.

    The widget also allows access and control over the player state including
    toggling of playing/paused and loop state, the current time, and the
    volume.

    Depending on the browser the video player supports mp4, webm, and ogg
    containers and a variety of codecs.

    Reference: https://panel.holoviz.org/reference/panes/Video.html

    :Example:

    >>> Video(
    ...     'https://file-examples-com.github.io/uploads/2017/04/file_example_MP4_640_3MG.mp4',
    ...     width=640, height=360, loop=True
    ... )
    """
    volume = param.Integer(default=100, bounds=(0, 100), doc="""
        The volume of the media player.""")

    _bokeh_model = _BkVideo

    _default_mime = 'mp4'

    _formats = ['mp4', 'webm', 'ogg']

    _media_type = 'video'
